import enum

class Vtype(enum.Enum):
    BINARY = 'binary'

    def __str__(self) -> str:
        return str(self.value)

class VariableVector:
    """
    Object That represent a vector of variable to be optimized.
    """
    def __init__(self, name='', size=1, vtype=Vtype.BINARY) -> None:
        if size < 1:
            raise ValueError("Variable vector size cannot be less than 1")

        self._name = name
        self._size = size
        self._vtype = vtype


    def size(self) -> int:
        """
        :return: size of this vector.
        """
        return self._size

    def vtype(self) -> Vtype:
        """
        :return: Type of variable in the vector.
        """
        return self._vtype

    def name(self) -> str:
        """
        :return: Name of this variable vector.
        """
        return self._name


from typing import Any, Dict, List, Tuple, Union
import numpy as np

class OptimizeResponse:
    """
    Object containing Optimization response and all its metrics.
    """
    def __init__(self, var_name: str, result_array: np.ndarray, metrics: Dict[str, Any]) -> None:
        self._var_name = var_name
        self._result_vector = result_array
        self._metrics = metrics

    def __getattr__(self, attr: str):
        # if attribute is the _var_name
        if attr == self._var_name:
            return self.result_vector()

        # else return the metrics name corresponding
        try:
            return self._metrics[attr]
        except KeyError as ex:
            raise AttributeError(attr) from ex


    def result_vector(self) -> np.ndarray:
        """
        :return: The result vector of this optimization.
        """
        return self._result_vector

    def result_items(self) -> List[Tuple[int, np.ndarray]]:
        """
        ex. [(-10000, [0, 1, 1, 0]), (-20000, [1, 0, 1, 0]), ...]

        :return: list of tuples containing the ising energy and it's corresponding result vector
        """
        return [(self.ising_energy[i], self._result_vector[i]) for i in range(len(self._result_vector))]

    def metrics(self, key: str = None) -> Union[str, Dict[str, Any]]:
        """
        :return: All metrics if no key is given of the specific metrics with the associated key if one is provided.

        :raise KeyError: The given key does not exist
        """
        if key:
            return self._metrics[key]
        else:
            return self._metrics



class TitanQError(Exception):
    """Base TitanQ error"""

class MaximumVariableLimitError(TitanQError):
    """Variable already defined in the model"""

class MissingVariableError(TitanQError):
    """Variable has not already been registered"""

class MissingObjectiveError(TitanQError):
    """Objective has not already been registered"""

class ObjectiveAlreadySetError(TitanQError):
    """An objective has already been set"""

class OptimizeError(TitanQError):
    """Error occur during optmization"""

class ServerError(TitanQError):
    """Error returned by the server"""


import datetime
import boto3
import botocore
import logging
import time

from .._client.model import AwsStorage, S3Input, S3Output
from .storage_client import StorageClient

log = logging.getLogger("TitanQ")


class S3Storage(StorageClient):
    def __init__(self, access_key, secret_key, bucket_name):
        """
        Initiate the s3 client for handling the temporary files.
        If any aws argument is missing this will raise an exception.

        :param aws_access_key_id: aws access key id used to upload and download files from an aws bucket.
        :param aws_secret_access_key: aws secret access key used to upload and download files from an aws bucket.
        :param aws_bucket_name: name of the aws bucket use to store temporairly data use by the titanq optimizer backend.
        """
        self._s3 = boto3.client('s3', aws_access_key_id=access_key, aws_secret_access_key=secret_key)
        self._access_key_id = access_key
        self._secret_access_key = secret_key
        self._bucket_name = bucket_name

        timestamp = datetime.datetime.now().isoformat()

        input_remote_folder = f"titanq_sdk_input/{timestamp}"
        self._bias_file_name = f"{input_remote_folder}/bias.npy"
        self._weights_file_name = f"{input_remote_folder}/weights.npy"\

        self._result_archive_file_name = f"titanq_sdk_output/{timestamp}/result.zip"

    def _upload_arrays(self, bias: bytes, weights: bytes):
        log.debug(f"Uploading object on AWS s3: {self._bias_file_name}")
        self._s3.put_object(Body=bias, Bucket=self._bucket_name, Key=self._bias_file_name)

        log.debug(f"Uploading object on AWS s3: {self._weights_file_name}")
        self._s3.put_object(Body=weights, Bucket=self._bucket_name, Key=self._weights_file_name)

    def _input(self) -> S3Input:
        return S3Input(
            weights_file_name=self._weights_file_name,
            bias_file_name=self._bias_file_name,
            s3=self._get_api_model_location())

    def _output(self) -> S3Output:
        return S3Output(
            result_archive_file_name=self._result_archive_file_name,
            s3=self._get_api_model_location())

    def _get_api_model_location(self) -> AwsStorage:
        """
        :return: An AwsStorage object that can be used in the api_model using this S3 credentials
        """
        return AwsStorage(
            bucket_name=self._bucket_name,
            access_key_id=self._access_key_id,
            secret_access_key=self._secret_access_key,
        )

    def wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        self._wait_for_file_to_be_uploaded(self._result_archive_file_name)
        return self._download_file(self._result_archive_file_name)

    def _wait_for_file_to_be_uploaded(self, filename: str):
        """
        Wait until a file exist in a bucket. It also verifies if the
        file is bigger than 0 bytes, this will ensure not downloading
        the empty archive file uploaded to test credentials

        :param filename: The full path of the file that is uploaded
        """
        log.debug(f"Waiting until object get upload on AWS s3: {filename}")
        while True:
            try:
                # check if file exist in the s3 bucket
                response = self._s3.head_object(
                    Bucket=self._bucket_name,
                    Key=filename,
                )
                # check if file content_length > 0
                if response['ContentLength'] > 0:
                    break
            except botocore.exceptions.ClientError as ex:
                # if the error we got is not 404, This is an unexpected error. Raise it
                if int(ex.response['Error']['Code']) != 404:
                    raise

            time.sleep(0.25) # wait 0.25 sec before trying again

    def _download_file(self, filename) -> bytes:
        """
        Download file from remote s3 bucket

        :param filename: The full path of the file to be uploaded

        :return: content of the file
        """
        log.debug(f"Downloading object from AWS s3: {filename}")
        object = self._s3.get_object(Bucket=self._bucket_name, Key=filename)
        return object['Body'].read()

    def _delete_remote_object(self):
        """
        Delete remote object on AWS s3

        :param key: object name to be deleted on the remote s3 bucket
        """
        for file_name in [self._bias_file_name, self._weights_file_name, self._result_archive_file_name]:
            log.debug(f"Deleting object on AWS s3: {file_name}")
            self._s3.delete_object(Bucket=self._bucket_name, Key=file_name)


import pytest
import numpy as np

from titanq import OptimizeResponse

def test_get_variable_result():
    result = np.random.rand(6).astype(np.float32)
    response = OptimizeResponse("x",  result, {})

    assert np.array_equal(response.x, result)
    assert np.array_equal(response.result_vector(), result)


def test_get_metrics():
    metrics = {"metrics1": 1, "metrics2": "value2"}
    response = OptimizeResponse("x",  np.random.rand(6).astype(np.float32), metrics)

    assert response.metrics1 == metrics["metrics1"]
    assert response.metrics2 == metrics["metrics2"]
    assert response.metrics("metrics1") == metrics["metrics1"]
    assert response.metrics("metrics2") == metrics["metrics2"]
    assert response.metrics() == metrics


import datetime
import io
import json
import numpy as np
import pytest
from typing import List
import uuid
import zipfile

from titanq import Model, Vtype, errors, Target
from titanq._client import  api_model
from titanq._storage.s3_storage import S3Storage
from titanq._storage.managed_storage import ManagedStorage

from .mock import TitanQClientMock, S3StorageMock


def file_in_filename_list(filename: str, filename_list: List[str]) -> bool:
    return any(filename in s for s in filename_list)


@pytest.fixture
def model_s3_client() -> Model:
    return Model(
        api_key="test_api_key",
        storage_client=S3Storage(
            access_key="aws_access_key",
            secret_key="aws_secret_access_key",
            bucket_name="bucket_name",
        )
    )

@pytest.fixture
def constant_datetime(monkeypatch):
    constant_datetime = datetime.datetime(2024,1,1,8,0,0)
    class MockDatetime(datetime.datetime):
        @classmethod
        def now(cls):
            return constant_datetime

    monkeypatch.setattr(datetime, 'datetime', MockDatetime)
    return constant_datetime

@pytest.mark.parametrize("api_key, storage_client ,expected_storage_class", [
    ("api_key", S3Storage(access_key="aws_access_key", secret_key="aws_secret_access_key", bucket_name="bucket_name"), S3Storage),
    ("api_key", ManagedStorage(TitanQClientMock()), ManagedStorage),
])
def test_selected_storage(api_key, storage_client, expected_storage_class):
    model = Model(api_key=api_key, storage_client=storage_client)
    assert isinstance(model._storage_client, expected_storage_class)


@pytest.mark.parametrize("name, size, vtype, error", [
    ('x', 1, Vtype.BINARY, None),
    ('x', 47, Vtype.BINARY, None),
    ('x', -1, Vtype.BINARY, ValueError),
    ('x', 0, Vtype.BINARY, ValueError)
])
def test_new_variable(model_s3_client, name, size, vtype, error):
    if error:
        with pytest.raises(error):
            model_s3_client.add_variable_vector(name, size, vtype)
    else:
        model_s3_client.add_variable_vector(name, size, vtype)


def test_multiple_variable(model_s3_client):
    model_s3_client.add_variable_vector('x', 1, Vtype.BINARY)

    with pytest.raises(errors.MaximumVariableLimitError):
        model_s3_client.add_variable_vector('y', 2, Vtype.BINARY)


@pytest.mark.parametrize("weights_shape, bias_shape, objective, error", [
    ((10, 10), (10,), Target.MINIMIZE, None),
    ((11, 10), (10,), Target.MINIMIZE, ValueError),
    ((10, 11), (10,), Target.MINIMIZE, ValueError),
    ((11, 11), (10,), Target.MINIMIZE, ValueError),
    ((10, 10, 10), (10,), Target.MINIMIZE, ValueError),
    ((10,), (10,), Target.MINIMIZE, ValueError),
    ((10,10), (9,), Target.MINIMIZE, ValueError),
    ((10,10), (10,1), Target.MINIMIZE, ValueError),
    ((10,10), (10,2), Target.MINIMIZE, ValueError),
])
def test_set_objective(model_s3_client, weights_shape, bias_shape, objective, error):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    weights = np.random.rand(*weights_shape).astype(np.float32)
    bias = np.random.rand(*bias_shape).astype(np.float32)

    if error:
        with pytest.raises(error):
            model_s3_client.set_objective_matrices(weights, bias, objective)
    else:
        model_s3_client.set_objective_matrices(weights, bias, objective)

@pytest.mark.parametrize("weights_data_type, bias_data_type, error", [
    (np.float32, np.float32, None),
    (np.float64, np.float32, ValueError),
    (np.float32, np.float64, ValueError),
    (np.int32, np.float32, ValueError),
    (np.float32, np.int32, ValueError),
    (np.bool_, np.float32, ValueError),
    (np.float32, np.bool_, ValueError),
    (np.byte, np.float32, ValueError),
    (np.float32, np.byte, ValueError),
    (np.short, np.float32, ValueError),
    (np.float32, np.short, ValueError),
])
def test_objective_matrices_data_type(model_s3_client, weights_data_type, bias_data_type, error):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    weights = np.random.rand(10, 10).astype(weights_data_type)
    bias = np.random.rand(10).astype(bias_data_type)

    if error:
        with pytest.raises(error):
            model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)
    else:
        model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)


def test_set_objective_without_variable(model_s3_client):
    weights = np.random.rand(10, 10)
    bias = np.random.rand(10)

    with pytest.raises(errors.MissingVariableError):
        model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)


def test_set_2_objective(model_s3_client):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    weights = np.random.rand(10, 10).astype(np.float32)
    bias = np.random.rand(10).astype(np.float32)

    model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)

    with pytest.raises(errors.ObjectiveAlreadySetError):
        model_s3_client.set_objective_matrices(weights, bias, Target.MINIMIZE)


def test_optimize_no_variable(model_s3_client):
    with pytest.raises(errors.MissingVariableError):
        model_s3_client.optimize()


def test_optimize_no_objective(model_s3_client):
    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)

    with pytest.raises(errors.MissingObjectiveError):
        model_s3_client.optimize()


def test_optimize(model_s3_client, constant_datetime):
    # expected npy result file content
    expected_result = np.random.rand(10).astype(np.float32)
    expected_result_buffer = io.BytesIO()
    np.save(expected_result_buffer, expected_result)

    # expected metrics file content
    expected_metrics = {"metrics1": 1, "metrics2": "value2"}

    # set mock client with mock values
    buff = io.BytesIO()
    with zipfile.ZipFile(buff, 'w') as file:
        file.writestr("result.npy", expected_result_buffer.getvalue())
        file.writestr("metrics.json", json.dumps(expected_metrics).encode())

    s3_mock = S3StorageMock([buff.getvalue()])
    model_s3_client._storage_client = s3_mock
    model_s3_client._titanq_client = TitanQClientMock(
        api_model.SolveResponse(computation_id=str(uuid.uuid4()), status="Queued", message="test_queued")
    )

    # optimize using sdk
    weights = np.random.rand(10, 10).astype(np.float32)
    bias = np.random.rand(10).astype(np.float32)

    model_s3_client.add_variable_vector('x', 10, Vtype.BINARY)
    model_s3_client.set_objective_matrices(weights, bias)

    response = model_s3_client.optimize()

    assert np.array_equal(response.x, expected_result)
    assert response.metrics() == expected_metrics

    # check that needed files has been uploaded
    assert file_in_filename_list("weights.npy", s3_mock.object_uploaded)
    assert file_in_filename_list("bias.npy", s3_mock.object_uploaded)

    # check that cleanup has been made
    assert file_in_filename_list(f"titanq_sdk_output/result_{constant_datetime.isoformat()}.zip", s3_mock.object_deleted)
    assert file_in_filename_list("weights.npy", s3_mock.object_deleted)
    assert file_in_filename_list("bias.npy", s3_mock.object_deleted)

    # check that result files have been downloaded
    assert file_in_filename_list(f"titanq_sdk_output/result_{constant_datetime.isoformat()}.zip", s3_mock.object_downloaded)


import datetime
import pytest
import requests
import requests_mock

from pydantic import ValidationError

from titanq._client import Client, api_model
from titanq import errors

@pytest.fixture
def solve_request():
    return api_model.SolveRequest(
        input=api_model.S3Input(
            bias_file_name='test_bias_file_name.npy',
            weights_file_name='test_weights_file_name.npy',
            s3=api_model.AwsStorage(
                bucket_name="test_input_bucket",
                access_key_id="test_access_key",
                secret_access_key="test_secret_access_key",
            ),
        ),
        output=api_model.S3Output(
                result_archive_file_name="test.zip",
                s3=api_model.AwsStorage(
                bucket_name="test_output_bucket",
                access_key_id="test_access_key",
                secret_access_key="test_secret_access_key",
            )
        ),
        parameters=api_model.Parameters(
            beta=[0.1],
            coupling_mult=0.5,
            num_chains=8,
            num_engines=2,
            timeout_in_secs=10,
            variables_format='binary'
        )
    )

@pytest.fixture
def baseurl():
    return "https://unittest.titanq.infinityq.io"

@pytest.fixture
def client(baseurl):
    return Client("test_api_key", baseurl)

@pytest.fixture
def temp_storage_raw_response():
    return \
        '{"input": {' \
            '"weights.npy": { "upload": "http://weights-upload.com", "download": "http://weights-download.com" },' \
            '"bias.npy": { "upload": "http://bias-upload.com", "download": "http://bias-download.com" } }, ' \
        '"output": {' \
            '"result.zip": { "upload": "http://result-upload.com", "download": "http://result-download.com" } } }'

def test_temp_storage(client, baseurl, temp_storage_raw_response):
    with requests_mock.Mocker() as m:
        adapter = m.get(f'{baseurl}/v1/temp_storage', content=temp_storage_raw_response.encode())
        temp_storage_response = client.temp_storage()

    assert adapter.call_count == 1
    assert adapter.last_request.headers['authorization'] == 'test_api_key'

    assert temp_storage_response.input.weights_file.upload == "http://weights-upload.com"
    assert temp_storage_response.input.weights_file.download == "http://weights-download.com"
    assert temp_storage_response.input.bias_file.upload == "http://bias-upload.com"
    assert temp_storage_response.input.bias_file.download == "http://bias-download.com"
    assert temp_storage_response.output.result_archive_file.upload == "http://result-upload.com"
    assert temp_storage_response.output.result_archive_file.download == "http://result-download.com"

def test_credits(client, baseurl):
    raw_credit_response = '{"remaining_credits": 10,"expiration_date": "2023-12-13T16:44:00+00:00"}'

    with requests_mock.Mocker() as m:
        adapter = m.get(f'{baseurl}/v1/credits', content=raw_credit_response.encode())
        credit_response = client.credits()

    assert adapter.call_count == 1
    assert adapter.last_request.headers['authorization'] == 'test_api_key'

    assert credit_response.expiration_date == datetime.datetime(2023, 12, 13, 16, 44, tzinfo=datetime.timezone.utc)
    assert credit_response.remaining_credits == 10


def test_credits_unexpected_error(client, baseurl):
    with requests_mock.Mocker() as m:
        m.get(f'{baseurl}/v1/credits', status_code=500)
        with pytest.raises(requests.exceptions.HTTPError):
            client.credits()

def test_solve_queued(client, baseurl, solve_request):
    raw_solve_response = '{ "computation_id": "test_id", "status": "Queued", "message": "test_message"}'
    with requests_mock.Mocker() as m:
        adapter = m.post(f'{baseurl}/v1/solve', content=raw_solve_response.encode())
        solve_response = client.solve(solve_request)

        assert adapter.call_count == 1
        assert adapter.last_request.headers['authorization'] == 'test_api_key'
        assert adapter.last_request.text == solve_request.model_dump_json()

        assert solve_response.computation_id == "test_id"
        assert solve_response.status == "Queued"
        assert solve_response.message == "test_message"


def test_solve_rejected(client, baseurl, solve_request):
    raw_solve_response = '{ "computation_id": "test_id", "status": "Rejected", "message": "test_message"}'
    with requests_mock.Mocker() as m:
        m.post(f'{baseurl}/v1/solve', content=raw_solve_response.encode())
        with pytest.raises(errors.ServerError):
            client.solve(solve_request)


def test_solve_validation_error(client, baseurl, solve_request):
    raw_solve_response = '{ "non_existant_field": "test_id", "status": "Queued", "message": "test_message"}'
    with requests_mock.Mocker() as m:
        m.post(f'{baseurl}/v1/solve', content=raw_solve_response.encode())
        with pytest.raises(ValidationError):
            client.solve(solve_request)


def test_solve_unexpected_error(client, baseurl, solve_request):
    with requests_mock.Mocker() as m:
        m.post(f'{baseurl}/v1/solve', status_code=500)
        with pytest.raises(requests.exceptions.HTTPError):
            client.solve(solve_request)



import logging
import requests
import time

from .._client.model import UrlInput, UrlOutput
from .storage_client import StorageClient
from .._client.client import Client

log = logging.getLogger("TitanQ")


class ManagedStorage(StorageClient):
    def __init__(self, titanq_client: Client):
        """
        Initiate the managed storage client for handling the temporary files.

        :titanq_client: titanq_client to be used to fetch temporary URL's
        """
        self._titanq_client = titanq_client
        self._urls = titanq_client.temp_storage()

    def _upload_arrays(self, bias: bytes, weights: bytes):
        log.debug(f"Uploading bias array on our temporary storage")
        requests.put(self._urls.input.bias_file.upload, data=bias)

        log.debug(f"Uploading weights array on our temporary storage")
        requests.put(self._urls.input.weights_file.upload, data=weights)

    def _input(self) -> UrlInput:
        return UrlInput(
            weights_file_name=self._urls.input.weights_file.download,
            bias_file_name=self._urls.input.bias_file.download)

    def _output(self) -> UrlOutput:
        return UrlOutput(result_archive_file_name=self._urls.output.result_archive_file.upload)

    def wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        self._wait_for_file_to_be_uploaded(self._urls.output.result_archive_file.download)
        return self._download_file(self._urls.output.result_archive_file.download)

    def _wait_for_file_to_be_uploaded(self, url: str):
        """
        Wait until the content of the file in the temporary storage is bigger
        than 0 bytes. Meaning it will wait until the file is uploaded

        :param url: Url to download the file.
        """
        response = requests.get(url)
        response.raise_for_status()
        while len(response.content) == 0:
            time.sleep(0.25)
            response = requests.get(url)
            response.raise_for_status()

    def _download_file(self, url) -> bytes:
        """
        Download file from the temporary storage

        :param url: Url to download the file.

        :return: content of the file in bytes
        """
        log.debug(f"Downloading object from the temporary storage")
        request = requests.get(url)
        return request.content

    def _delete_remote_object(self) -> None:
        log.debug("Temporary storage option does not delete any file at the moment")


from datetime import datetime
from pydantic import BaseModel, Field, field_serializer, SecretStr
from typing import List, Union


#########################
##    solve models     ##
#########################
class AwsStorage(BaseModel):
    """
    S3 backend storage
    """
    bucket_name: str
    access_key_id: SecretStr
    secret_access_key: SecretStr

    @field_serializer('access_key_id', 'secret_access_key', when_used='json')
    def dump_secret(self, v):
        return v.get_secret_value()


class S3Input(BaseModel):
    """
    Input object of the solve request with s3
    """
    bias_file_name: str
    weights_file_name: str
    s3: AwsStorage


class UrlInput(BaseModel):
    """
    Input object of the solve request with url
    """
    # always true, the user does not have to set this value
    file_name_is_url: bool = Field(default=True, frozen=True)
    bias_file_name: str
    weights_file_name: str


class S3Output(BaseModel):
    """
    Output object of the solve request with s3
    """
    result_archive_file_name: str
    s3: AwsStorage

class UrlOutput(BaseModel):
    """
    Output object of the solve request with url
    """
    # always true, the user does not have to set this value
    file_name_is_url: bool = Field(default=True, frozen=True)
    result_archive_file_name: str


class Parameters(BaseModel):
    """
    Tuning parameters used by the solver
    """
    beta: List[float]
    coupling_mult: float
    timeout_in_secs: float
    num_chains: int
    num_engines: int
    variables_format: str


class SolveRequest(BaseModel):
    """
    The actual solve request object send to the backend
    """
    input: Union[S3Input, UrlInput]
    output: Union[S3Output, UrlOutput]
    parameters: Parameters


class SolveResponse(BaseModel):
    """
    The response object returned by the backend on solve request
    """
    computation_id: str
    status: str
    message: str


#########################
##   credits models    ##
#########################
class CreditsResponse(BaseModel):
    """
    The response object returned by the backend on credits request
    """
    remaining_credits: int
    expiration_date: datetime



#########################
## temp_storage models ##
#########################
class FileUrls(BaseModel):
    """
    Object always containing a pair of a download and an upload url
    """
    download: str
    upload: str


class TempStorageInput(BaseModel):
    """
    Object containing input files for temporary storage object
    """
    weights_file: FileUrls = Field(alias="weights.npy")
    bias_file: FileUrls = Field(alias="bias.npy")


class TempStorageOutput(BaseModel):
    """"
    Object containing output files for temporary storage object
    """
    result_archive_file: FileUrls = Field(alias="result.zip")


class TempStorageResponse(BaseModel):
    """
    The response object returned by the backend on temporary storage response
    """
    input: TempStorageInput
    output: TempStorageOutput


import io
import json
import logging
import numpy as np
import zipfile
from typing import Any, Dict, List, Tuple

from .errors import MaximumVariableLimitError, MissingVariableError, MissingObjectiveError, ObjectiveAlreadySetError, OptimizeError
from .objective import Objective, Target
from .optimize_response import OptimizeResponse
from .variable import VariableVector, Vtype
from .._client import api_model, Client
from .._storage import ManagedStorage, StorageClient

log = logging.getLogger("TitanQ")

_TITANQ_BASE_URL = "https://titanq.infinityq.io"


class Model:
    """
    Root object to define a problem to be optimize
    """

    def __init__(
        self,
        *,
        api_key: str,
        storage_client: StorageClient = None,
        base_server_url: str = _TITANQ_BASE_URL
        ) -> None:
        """
        Initiate the model with a storage client depending on the provided arguments.

        If the storage_client is missing, the storage will be managed by titanQ.

        NOTE: the storage manged by titanQ supports weight matrices with a size up to 10k

        :param api_key: TitanQ api key to access the service.
        :param storage_client: Storage to choose in order to store some items
        :param base_server_url: titanQ server url


        Example
        ~~~~~~~
        ..  highlight:: python
        ..  code-block:: python
        from titanq import Model, S3Storage

        # s3 storage on AWS
        model = Model(
            api_key="{insert API key here}",
            storage_client=S3Storage(
                access_key="{insert aws bucket access key here}",
                secret_key="{insert aws bucket secret key here}",
                bucket_name="{insert bucket name here}"
            )
        )

        # managed Storage (Up to 10k variables only)
        model = Model(api_key="{insert API key here}")
        """
        self._variables: VariableVector = None
        self._objective: Objective = None
        self._titanq_client = Client(api_key, base_server_url)

        # the user chose a managed storage or left it as default
        if storage_client is None:
            storage_client = ManagedStorage(self._titanq_client)

        self._storage_client = storage_client


    def add_variable_vector(self, name='', size=1, vtype=Vtype.BINARY):
        """
        Add a vector of variable to the model. Only a single vector of variables can be set.
        Calling this method again will override the current vector of variables

        :param name: The name given to this variable vector.
        :param size: The size of the vector.
        :param vtype: Type of the variables inside the vector.

        :raise MaximumVariableLimitError: If a variable is already defined.
        :raise ValueError: If the size of the vector is < 1.

        Example
        ~~~~~~~
        ..  highlight:: python
        ..  code-block:: python
        import numpy as np
        from titanq import Model, Vtype

        # set up model
        ...

        model.add_variable_vector('x', size, Vtype.BINARY)
        """
        if self._variables is not None:
            raise MaximumVariableLimitError("Cannot add additional variable without busting the maximum number of variable (1).")

        log.debug(f"add variable name='{name}'.")

        self._variables = VariableVector(name, size, vtype)


    def set_objective_matrices(self, weights: np.ndarray, bias: np.ndarray, target=Target.MINIMIZE):
        """
        Set the objective matrices for the model.

        :param weights: The quadratic objective matrix, **this matrix needs to be symmetrical**
        :type weights: a NumPy 2-D dense ndarray (must be float32)

        :param bias: The linear constraint vector
        :type bias:  a NumPy 1-D ndarray

        :param target: The target of this objective matrix.
        :type target: Target Enum

        :raise MissingVariableError: If no variable have been added to the model.
        :raise ObjectiveAlreadySetError: If an objective has already been setted in this model.
        :raise ValueError: If the weights shape or the bias shape does not fit the variable in the model.
        :raise ValueError: If the weights or bias data type is not f32.

        Example
        ~~~~~~~
        ..  highlight:: python
        ..  code-block:: python
        from titanq import Model, Target

        edges = {0:[4,5,6,7], 1:[4,5,6,7], 2:[4,5,6,7], 3:[4,5,6,7], 4:[0,1,2,3], 5:[0,1,2,3], 6:[0,1,2,3], 7:[0,1,2,3]}
        size = len(edges)

        # construct the weight matrix from the edges list
        weights = np.zeros((size, size), dtype=np.float32)
        for root, connections in edges.items():
            for c in connections:
                weights[root][c] = 1

        # construct the bias vector (Uniform weighting across all nodes)
        bias = np.asarray([0]*size, dtype=np.float32)

        # set up model
        ...

        model.set_objective_matrices(weights, bias, Target.MINIMIZE)
        """

        if self._variables is None:
            raise MissingVariableError("Cannot set objective before adding a variable to the model.")

        if self._objective is not None:
            raise ObjectiveAlreadySetError("An objective has already have been set for this model.")

        log.debug(f"set objective matrix and bias vector.")

        self._objective = Objective(self._variables.size(), weights, bias, target)


    def optimize(self, *, beta=[0.1], coupling_mult=0.5, timeout_in_secs=10.0, num_chains=8, num_engines=1):
        """
        Optimize this model.

        Note: All of the files used during the computation will be cleaned at the end.

        :param beta: beta hyper parameter used by the backend solver.
        :param coupling_mult: coupling_mult hyper parameter used by the backend solver.
        :param timeout_in_secs: Maximum time (in seconds) the backend solver can take to resolve this problem.
        :param num_chains: num_chains hyper parameter used by the backend solver.
        :param num_engines: num_engines parameter used by the backend solver.

        :raise MissingVariableError: If no variable have been added to the model.
        :raise MissingObjectiveError: If no variable have been added to the model.

        Example
        ~~~~~~~
        ..  highlight:: python
        ..  code-block:: python
        # set up model, objective and variable
        ...

        # basic solve
        response = model.optimize(timeout_in_secs=60)

        # multiple engine
        response = model.optimize(timeout_in_secs=60, num_engines=2)

        # custom values
        response = model.optimize(beta=[0.1], coupling_mult=0.75, num_chains=8)

        # print values
        print("-" * 15, "+", "-" * 26, sep="")
        print("Ising energy   | Result vector")
        print("-" * 15, "+", "-" * 26, sep="")
        for ising_energy, result_vector in response.result_items():
            print(f"{ising_energy: <14f} | {result_vector}")
        """
        if self._variables is None:
            raise MissingVariableError("Cannot optimize before adding a variable to the model.")

        if self._objective is None:
            raise MissingObjectiveError("Cannot optimize before adding an objective to the model.")

        result, metrics = self._solve(beta, coupling_mult, timeout_in_secs, num_chains, num_engines)

        return OptimizeResponse(self._variables.name(), result, metrics)

    def _solve(self,
            beta: List[float],
            coupling_mult: float,
            timeout_in_secs: float,
            num_chains: int,
            num_engines: int
        ) -> Tuple[np.ndarray, Dict[str, Any]]:
        """
        issue a solve request and wait for it to complete.

        :param beta: beta hyper parameter used by the backend solver.
        :param coupling_mult: coupling_mult hyper parameter used by the backend solver.
        :param timeout_in_secs: Maximum time (in seconds) the backend solver can take to resolve this problem.
        :param num_chains: num_chains hyper parameter used by the backend solver.
        :param num_engines: num_engines parameter used by the backend solver.

        :return: the result numpy array and the metric json object
        """
        with self._storage_client.temp_files_manager(
            self._objective.bias(),
            self._objective.weights()
        ) as temp_files:
            request = api_model.SolveRequest(
                input=temp_files.input(),
                output=temp_files.output(),
                parameters=api_model.Parameters(
                    beta=beta,
                    coupling_mult=coupling_mult,
                    num_chains=num_chains,
                    num_engines=num_engines,
                    timeout_in_secs=timeout_in_secs,
                    variables_format=str(self._variables.vtype())
                )
            )

            solve_response = self._titanq_client.solve(request)

            # wait for result to be uploaded by the solver and download it
            archive_file_content = temp_files.download_result()
            with zipfile.ZipFile(io.BytesIO(archive_file_content), 'r') as zip_file:
                try:
                    metrics_content = zip_file.read("metrics.json")
                    result_content = zip_file.read("result.npy")
                except KeyError as ex:
                    raise OptimizeError(
                        "Unexpected error in the solver, please contact titanQ support for more help" \
                        f" and provide the following computation id {solve_response.computation_id}") from ex

        log.debug("Optimization completed")
        return np.load(io.BytesIO(result_content)), json.loads(metrics_content)



import numpy as np

from abc import ABC, abstractmethod
from io import BytesIO
from typing import Type, Union


from .._client.model import S3Input, S3Output, UrlInput, UrlOutput


class StorageClient(ABC):
    def __init__(self):
        pass

    def temp_files_manager(self, bias: np.ndarray, weights: np.ndarray) -> Type["TempFileManager"]:
        """'
        Creates and returns a context manager that will be handling the temporary file management

        :param bias: bias numpy array
        :param weights: weights numpy array
        """
        return TempFileManager(self, bias, weights)

    @abstractmethod
    def _input(self) -> Union[S3Input, UrlInput]:
        """
        Returns the api model for the input of the solve request

        :return: either the s3 or the url input
        """

    @abstractmethod
    def _output(self) -> Union[S3Output, UrlOutput]:
        """
        Returns the api model for the output of the solve request

        :return: either the s3 or the url output
        """

    @abstractmethod
    def _upload_arrays(self, bias: bytes, weights: bytes):
        """
        Uploads .npy arrays (bias and weights) to the storage client

        :param bias: bias array in bytes
        :param weights: weights array in bytes
        """

    @abstractmethod
    def wait_for_result_to_be_uploaded_and_download(self) -> bytes:
        """
        Wait until a file is uploaded on the storage client and download it

        :return: content of the result file in bytes
        """

    @abstractmethod
    def _delete_remote_object(self) -> None:
        """
        Deletes a remote object on the storage client
        """

class TempFileManager():
    """
    Context manager handling the upload and download to a temporary storage
    """
    def __init__(
        self,
        storage_client: StorageClient,
        bias: np.ndarray,
        weights: np.ndarray
    ) -> None:
        """
        :param storage_client: storage client that will handle uploads, downloads and generating api models
        :param bias: bias numpy array
        :param weights: weights numpy array
        """

        super().__init__()
        self._storage_client = storage_client
        self._bias = bias
        self._weights = weights

    def __enter__(self) -> Type["TempFileManager"]:
        self._storage_client._upload_arrays(
            self._to_bytes(self._bias),
            self._to_bytes(self._weights))
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self._storage_client._delete_remote_object()

    def input(self) -> Union[S3Input, UrlInput]:
        return self._storage_client._input()

    def output(self) -> Union[S3Output, UrlOutput]:
        return self._storage_client._output()

    def download_result(self):
        return self._storage_client.wait_for_result_to_be_uploaded_and_download()

    def _to_bytes(self, array: np.ndarray) -> bytes:
        """
        :return: numpy array as bytes
        """
        buffer = BytesIO()
        np.save(buffer, array)
        return buffer.getvalue()